#define SEND_HSV 1
//#define DISABLE_COLOR 1

#include <LiquidCrystal_I2C.h>

#include <Servo.h>

#include <Adafruit_TCS34725.h>


#include <ros.h>
#include <std_msgs/Bool.h>
#include <std_msgs/Int32.h>
#include <colors_msgs/Colors.h>
#include <ctrlpan_msgs/alim_stat.h>

#ifdef SEND_HSV
  #include <geometry_msgs/Vector3.h>
#endif


#define COL_L 4
#define COL_R 7
Adafruit_TCS34725 tcs = Adafruit_TCS34725(TCS34725_INTEGRATIONTIME_24MS, TCS34725_GAIN_1X);
ros::NodeHandle node_handle;
colors_msgs::Colors color_msg;

#ifdef SEND_HSV
  geometry_msgs::Vector3 hsv;
#endif

// Servo object declaration
Servo flagServo; 
#define FLAG_UNFOLDED_ANGLE 195
#define FLAG_FOLDED_ANGLE 20

Servo armServo; //pour manche à air
#define ARM_UNFOLDED_ANGLE 90
#define ARM_FOLDED_ANGLE 180

// Pin declaration
const int flagServoPin = D3;
const int armServoPin = D2;

const int pullCordPin = A6;
const int switchPin = D6;
const int buttonPin = D12;
const int buttonLedPin = D11;

const int voltage_twelvePin = A3;
const int voltage_fivePin = A0;
const int current_twelvePin = A2;
const int current_fivePin = A1;
#define VOLT12_RATIO 1
#define VOLT5_RATIO 1
#define CURR5_RATIO 1
#define CURR12_RATIO 1

const int bauStatusPin = D8;
const int bauSoftPin = D9;


// Flag state
bool flagStatus;

// BAU state
bool bauSoftStatus;
bool bauStatus;

// Arm state
bool armStatus;



// Current team (either TEAM_YELLOW or TEAM_BLUE)
bool team;
bool buttonStatus;
bool buttonLedState;

// Last received estimated score
int estimatedScore = 0;

/**
 * Tirette (avec gestion du rebond)
 */

int pullCordState;            // the current reading from the input pin
int lastPullCordStatus = LOW; // the previous reading from the input pin

unsigned long lastDebounceTime = 0; // the last time the output pin was toggled
unsigned long debounceDelay = 50;   // the debounce time; increase if the output flickers

unsigned long lastMessageTime = 0;
unsigned int const MSG_SENDING_INTERVAL = 100;

ros::NodeHandle nh;

std_msgs::Bool teamMsg;
std_msgs::Bool buttonStatusMsg;
std_msgs::Bool pullCordStatusMsg;
std_msgs::Bool bauStatusMsg;
ctrlpan_msgs::alim_stat alimStatMsg;

ros::Publisher teamStatusPub("team", &teamMsg);
ros::Publisher buttonStatusPub("button_status", &buttonStatusMsg);
ros::Publisher bauStatusPub("BAU_status", &bauStatusMsg);
ros::Publisher pullCordStatusPub("pull_cord_status", &pullCordStatusMsg);
ros::Publisher colorSensorsPub("color_sensors", &color_msg);
ros::Publisher alimStatPub("alim_stat", &alimStatMsg);


#ifdef SEND_HSV
ros::Publisher colorSensorsPub2("hsv_left", &hsv);
ros::Publisher colorSensorsPub3("hsv_right", &hsv);
#endif




void TCA9548A(uint8_t bus)
{
  Wire.beginTransmission(0x70);  // TCA9548A address is 0x70
  Wire.write(1 << bus);          // send byte to select bus
  Wire.endTransmission();
}



void estimatedScoreCallback(const std_msgs::Int32 &estimatedScore);
ros::Subscriber<std_msgs::Int32> estimatedScoreSub("estimated_score", &estimatedScoreCallback);

void flagStatusChange(const std_msgs::Bool &flagStatusMsg);
ros::Subscriber<std_msgs::Bool> flagStatusSub("flag_state", &flagStatusChange);

void armStatusChange(const std_msgs::Bool &armStatusMsg);
ros::Subscriber<std_msgs::Bool> armStateSub("arm_state", &armStatusChange);

void bauSoftSet(const std_msgs::Bool &bauSoftSetMsg);
ros::Subscriber<std_msgs::Bool> bauSoftSub("bau_state", &bauSoftSet);

void buttonLedSet(const std_msgs::Bool &buttonLedSetMsg);
ros::Subscriber<std_msgs::Bool> buttonLedSub("button_led_state", &buttonLedSet);


// LCD screen
LiquidCrystal_I2C lcd(0x27, 16, 2);

void refreshScreen()
{
  //lcd.init();
  //lcd.backlight();
  
  lcd.clear();

  lcd.setCursor(0, 0);
  lcd.print("Score=");
  lcd.setCursor(6, 0);
  lcd.print(estimatedScore);

  lcd.setCursor(0, 1);
  lcd.print(team ? "Team=JAUNE" : "Team=BLEU");
  lcd.setCursor(12, 1);
  lcd.print(pullCordState ? "GO" : "STOP");
  lcd.setCursor(13, 0);
  lcd.print(buttonStatus ? "ON" : "OFF");
}

void setupColorSensors(){

  TCA9548A(COL_L);

  if (tcs.begin()) {
    nh.loginfo("Found left color sensor");
  } else {
    nh.logerror("Left sensor not found");
  }


  TCA9548A(COL_R);

  if (tcs.begin()) {
    nh.loginfo("Found right color sensor");
  } else {
    nh.logerror("Right sensor not found");
  }
  
  nh.advertise(colorSensorsPub);
  nh.advertise(colorSensorsPub2); // TO REMOVE
  nh.advertise(colorSensorsPub3); // TO REMOVE

}

void setup()
{

  // INIT Wire on : A4 => SDA  ||  A5 => SCL
	Wire.begin();

  flagServo.attach(flagServoPin);
  flagServo.write(FLAG_FOLDED_ANGLE);

  armServo.attach(armServoPin);
  armServo.write(ARM_FOLDED_ANGLE);

  pinMode(pullCordPin, INPUT_PULLUP);
  pinMode(switchPin, INPUT_PULLUP);
  pinMode(buttonPin, INPUT_PULLUP);
  pinMode(buttonLedPin, OUTPUT);

  pinMode(bauSoftPin, OUTPUT);
  pinMode(bauStatusPin, INPUT);

  digitalWrite(buttonLedPin, HIGH);

  lcd.init();
  lcd.backlight();

  team = digitalRead(switchPin);
  buttonStatus = digitalRead(buttonPin);

  nh.initNode();

  nh.advertise(teamStatusPub);
  nh.advertise(buttonStatusPub);
  nh.advertise(pullCordStatusPub);
  nh.advertise(bauStatusPub);
  nh.advertise(alimStatPub);

  nh.subscribe(estimatedScoreSub);
  nh.subscribe(flagStatusSub);
  nh.subscribe(armStateSub);
  nh.subscribe(bauSoftSub);
  nh.subscribe(buttonLedSub);

  // color Sensors setup
  #ifndef DISABLE_COLOR
    setupColorSensors();
  #endif

  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Waiting ROS...");

  while (!nh.connected())
  {
    nh.spinOnce();
    delay(MSG_SENDING_INTERVAL);
  }

  refreshScreen();
}

// Appelé quand le score est mis à jour (message de score reçu sur l'Arduino depuis la raspi)
void estimatedScoreCallback(const std_msgs::Int32 &estimatedScoreMsg)
{
  estimatedScore = estimatedScoreMsg.data;
  refreshScreen();
}

// Appelé quand l'état du drapeau est mis à jour
void flagStatusChange(const std_msgs::Bool &flagStatusMsg)
{
  flagStatus = flagStatusMsg.data;
}

// Appelé quand l'état du BAU est mis à jour
void bauSoftSet(const std_msgs::Bool &bauSoftSetMsg)
{
    bauSoftStatus = bauSoftSetMsg.data;
}

// Appelé quand l'état du bras (manche à air) est mis à jour
void armStatusChange(const std_msgs::Bool &armStatusMsg)
{
    armStatus = armStatusMsg.data;
}

// Appelé quand l'état de la led du bouton change
void buttonLedSet(const std_msgs::Bool &buttonLedSetMsg)
{
    buttonLedState = buttonLedSetMsg.data;
}

void manage_supply_card_data(){

  int volt12_valueADC = analogRead(voltage_twelvePin);
  int volt5_valueADC = analogRead(voltage_fivePin);
  int curr5_valueADC = analogRead(current_fivePin);
  int curr12_valueADC = analogRead(current_twelvePin);

  float volt12_value = (float) volt12_valueADC * VOLT12_RATIO;
  float volt5_value = (float) volt5_valueADC * VOLT5_RATIO;
  float curr5_value = (float) curr5_valueADC * CURR5_RATIO;
  float curr12_value = (float) curr12_valueADC * CURR12_RATIO;

  alimStatMsg.voltage_12 = volt12_value;
  alimStatMsg.voltage_5 = volt5_value;
  alimStatMsg.current_12 = curr12_value;
  alimStatMsg.current_5 = curr5_value;

}

float max(float a, float b, float c) {
   return ((a > b)? (a > c ? a : c) : (b > c ? b : c));
}
float min(float a, float b, float c) {
   return ((a < b)? (a < c ? a : c) : (b < c ? b : c));
}

int rgb_to_hsv(float& h, float& s,float& v, uint16_t re,uint16_t gr,uint16_t bl){


   float r = ((float) re) / (255.0);
   float g = ((float) gr) / (255.0);
   float b = ((float) bl) / (255.0);

   float cmax = max(r, g, b); // maximum of r, g, b
   float cmin = min(r, g, b); // minimum of r, g, b
   float diff = cmax-cmin; // diff of cmax and cmin.

   if (cmax == cmin)
      h = 0;
   else if (cmax == r)
      h = fmod((60 * ((g - b) / diff) + 360), 360.0);
   else if (cmax == g)
      h = fmod((60 * ((b - r) / diff) + 120), 360.0);
   else if (cmax == b)
      h = fmod((60 * ((r - g) / diff) + 240), 360.0);
   // if cmax equal zero
      if (cmax == 0)
         s = 0;
      else
         s = (diff / cmax) * 100;
   // compute v
   v = cmax * 100;

   return 0;
}

int traduceHSVtoInt(float h, float s, float v){

  //WHITE
  if (v < 6){
    return color_msg.BLACK;
  }

  if (s < 18){
    return color_msg.WHITE;
  }

  if (h > 350 || h < 10){
    return color_msg.RED;
  }

  if (h > 20 && h < 40){
    return color_msg.ORANGE;
  }

  if (h > 100 && h < 130){
    return color_msg.GREEN;
  }

  return color_msg.UNKNOWN;
}

void colorSensorsLoop(){
  uint16_t r, g, b, c;
  float h, s, v;

  TCA9548A(COL_L);

  tcs.getRawData(&r, &g, &b, &c);
  rgb_to_hsv(h, s, v, r, g, b);
  int leftColor = traduceHSVtoInt(h, s, v);

#ifdef SEND_HSV   
    hsv.x = h;
    hsv.y = s;
    hsv.z = v;

    colorSensorsPub2.publish( &hsv ); // to remove
#endif

  TCA9548A(COL_R);

  tcs.getRawData(&r, &g, &b, &c);
  rgb_to_hsv(h, s, v, r, g, b);
  int rightColor = traduceHSVtoInt(h, s, v);
   
#ifdef SEND_HSV   
    hsv.x = h;
    hsv.y = s;
    hsv.z = v;

    colorSensorsPub3.publish( &hsv ); // to remove
#endif


  color_msg.color_left = leftColor;
  color_msg.color_right = rightColor;
  
  colorSensorsPub.publish( &color_msg );

  nh.spinOnce();
}

void loop()
{

  if (!nh.connected())
  {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Waiting ROS...");

    while (!nh.connected())
    {
      nh.spinOnce();
      delay(MSG_SENDING_INTERVAL);
    }

    refreshScreen();
  }


  // color Sensors loop
  #ifndef DISABLE_COLOR
    colorSensorsLoop();
  #endif
  // DEBUT PARTIE TIRETTE

  int pullCordReading = digitalRead(pullCordPin);

  // If the switch changed, due to noise or pressing:
  if (pullCordReading != lastPullCordStatus)
  {
    // reset the debouncing timer
    lastDebounceTime = millis();
  }

  if ((millis() - lastDebounceTime) > debounceDelay)
  {
    // whatever the reading is at, it's been there for longer than the debounce
    // delay, so take it as the actual current state:

    // if the button state has changed:
    if (pullCordReading != pullCordState)
    {
      pullCordState = pullCordReading;
      refreshScreen();
    }
  }

  lastPullCordStatus = pullCordReading;

  // FIN PARTIE TIRETTE

  if (!digitalRead(buttonPin) != buttonStatus)
  {
    // Appui bouton
    buttonStatus = !digitalRead(buttonPin);
    refreshScreen();
  }


  if (digitalRead(switchPin) != team)
  {
    //Changement d'équipe
    team = digitalRead(switchPin);
    refreshScreen();
  }


  bauStatus = digitalRead(bauStatusPin);

  if (millis() - lastMessageTime >= MSG_SENDING_INTERVAL)
  {
    teamMsg.data = team;
    teamStatusPub.publish(&teamMsg);

    pullCordStatusMsg.data = pullCordState;
    pullCordStatusPub.publish(&pullCordStatusMsg);

    buttonStatusMsg.data = buttonStatus;
    buttonStatusPub.publish(&buttonStatusMsg);

    bauStatusMsg.data = bauStatus;
    bauStatusPub.publish(&bauStatusMsg);

    manage_supply_card_data();
    alimStatPub.publish(&alimStatMsg);

    lastMessageTime = millis();
    nh.spinOnce();
  }

  if(flagStatus)
  {
    flagServo.write(FLAG_UNFOLDED_ANGLE);
  } else {
    flagServo.write(FLAG_FOLDED_ANGLE);
  }

  if(armStatus)
  {
    armServo.write(ARM_UNFOLDED_ANGLE);
  } else {
    armServo.write(ARM_FOLDED_ANGLE);
  }

  if(buttonLedState)
  {
    digitalWrite(buttonLedPin, HIGH);
  } else {
    digitalWrite(buttonLedPin, LOW);
  }

  digitalWrite(bauSoftPin, bauSoftStatus);
}
